# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import hashlib
from datetime import datetime, timezone
from urllib.parse import quote, urlparse, parse_qs

import requests

from .. import get_version
from ..config import config

from .auth import UHV1Signature


class HTTPError(requests.RequestException):
    """A generic error for HTTP requests."""


class Request:

    # pylint: disable=too-many-arguments
    def __init__(self, url, method, payload='', json=False,
                 headers=None, **requests_kwargs):
        self.url = url
        self._url = urlparse(self.url)
        self.method = method.upper()
        self.payload = payload
        self._requests_kwargs = requests_kwargs

        self.date = datetime.now(timezone.utc)
        self.payload_sha256 = self._generate_payload_sha256()

        headers = headers if headers else {}
        self.headers = {
            'User-Agent': 'updatehub-utils/{}'.format(get_version()),
            'Host': self._url.netloc,
            'Timestamp': self.date.timestamp(),
            'Content-sha256': self.payload_sha256,
            'Api-Content-Type': 'application/vnd.updatehub-v1+json',
            'Accept': 'application/json',
        }
        self.headers.update(headers)
        if json:
            self.headers['Content-Type'] = 'application/json'

    def _generate_payload_sha256(self):
        payload = self.payload
        if not isinstance(payload, bytes):
            payload = payload.encode()
        return hashlib.sha256(payload).hexdigest()

    def _canonical_query(self):
        raw_query = parse_qs(self._url.query)
        query = []
        for key in raw_query:
            for value in sorted(raw_query[key]):
                query.append('{}={}'.format(key, quote(value)))
        return '&'.join(sorted(query))

    def _canonical_headers(self):
        normalized_headers = [(k.strip().lower(), str(v).strip())
                              for k, v in self.headers.items()]
        headers = ['{}:{}'.format(header, value)
                   for header, value in normalized_headers]
        return '\n'.join(sorted(headers))

    def canonical(self):
        request = '{method}\n{uri}\n{query}\n{headers}\n\n{payload}'
        return request.format(
            method=self.method,
            uri=self._url.path,
            query=self._canonical_query(),
            headers=self._canonical_headers(),
            payload=self.payload_sha256,
        )

    def _sign(self):
        try:
            access, secret = config.get_credentials()
        except ValueError:
            raise HTTPError('Could not sign request. Missing credentials.')
        signature = UHV1Signature(self, access, secret)
        self.headers['Authorization'] = signature.signature

    def _prepare_headers(self):
        """Transforms all header values in strings.

        This must be called before any real HTTP request to prevent
        breaking any library that just expects strings as header
        values.
        """
        return {header: str(self.headers[header]) for header in self.headers}

    def send(self):
        self._sign()
        headers = self._prepare_headers()
        response = requests.request(
            self.method,
            self.url,
            headers=headers,
            data=self.payload,
            timeout=30,
            **self._requests_kwargs
        )
        return response



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import sys
from contextlib import contextmanager

from ..core.package import Package
from ..core.utils import dump_package, load_package
from ..utils import get_local_config_file
from ..ui import show_cursor


@contextmanager
def open_package(read_only=False):
    """Context manager for package operations.

    It opens a package, gives control to the user and, finally, dumps
    the package. If read_only, it does not dump the package.
    """
    pkg_file = get_local_config_file()
    try:
        package = load_package(pkg_file)
    except FileNotFoundError:
        package = Package()
    except ValueError as err:
        print('Invalid configuration file: {}'.format(err))
        sys.exit(1)
    yield package
    if not read_only:
        dump_package(package.to_template(), pkg_file)


def error(code, msg):
    """Terminates cli with an error code and message for the user."""
    print('Error: {}'.format(msg))
    show_cursor()
    sys.exit(code)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0
"""UHU REPL helper functions.

Includes reusable prompts, auto-completers, constraint checkers.
"""

import sys
from functools import partial, wraps

from prompt_toolkit import prompt
from prompt_toolkit.contrib.completers import WordCompleter
from prompt_toolkit.filters import HasCompletions
from prompt_toolkit.key_binding.manager import KeyBindingManager
from prompt_toolkit.keys import Keys

from ..core.validators import validate_option_requirements
from ..core.object import Modes
from ..core._options import Options

from .completers import (
    ObjectFilenameCompleter, ObjectModeCompleter, ObjectOptionValueCompleter,
    ObjectUIDCompleter, YesNoCompleter)
from .exceptions import CancelPromptException
from .validators import (
    ObjectUIDValidator, ContainerValidator, ObjectOptionValueValidator,
    PackageUIDValidator)


manager = KeyBindingManager.for_prompt()  # pylint: disable=invalid-name


@manager.registry.add_binding(Keys.ControlD)
def ctrl_d(_):
    """Ctrl D quits appliaction returning 0 to sys."""
    sys.exit(0)


@manager.registry.add_binding(Keys.ControlC)
def ctrl_c(_):
    """Ctrl C raises an exception to be caught by functions.

    Main prompt must exit uhu with code status 1, while subprompts
    must returns to main prompt.
    """
    raise CancelPromptException('Cancelled operation.')


@manager.registry.add_binding(Keys.Enter, filter=HasCompletions())
def enter(event):
    """Enter selects a completion when has completions.

    When there is no completion available, submit value.
    """
    buffer = event.current_buffer
    state = buffer.complete_state
    if len(state.current_completions) == 1:
        state = state.go_to_index(0)
        buffer.apply_completion(state.current_completion)
    elif state.current_completion is None:
        buffer.complete_next()
    else:
        buffer.apply_completion(buffer.complete_state.current_completion)


def cancellable(func):
    """Decorator to cancell a current prompt."""
    @wraps(func)
    def wrapper(*args, **kw):
        try:
            return func(*args, **kw)
        except CancelPromptException:
            pass  # Do nothing cancelling the current command
    return wrapper


# pylint: disable=invalid-name
prompt = partial(prompt, key_bindings_registry=manager.registry)


def check_arg(ctx, msg):
    """Checks if user has passed an argument.

    :param msg: The error message to display to the user in when an
                argument is not passed.
    """
    if ctx.arg is None:
        raise ValueError(msg)


def check_version(ctx):
    """Checks if package already has a version."""
    if ctx.package.version is None:
        raise ValueError('You need to set a version first')


def check_product(ctx):
    """Checks if product is already set."""
    if ctx.package.product is None:
        raise ValueError('You need to set a product first')


def set_product_prompt(product):
    """Sets prompt to be 'uhu [product]'."""
    return '[{}] uhu> '.format(product[:6])


def parse_prompt_object_uid(value):
    """Parses value passed to a prompt using get_objects_completer.

    :param value: A value returned by :func:`get_objects_completer`.
    """
    return int(value.split('#')[0].strip())


def prompt_object_options(n_sets, object_mode):
    """Prompts user for object options.

    :param n_sets: The number of installation sets in package.
    :param object_mode: A string indicating the object mode.
    """
    values = {}
    mode = Modes.get(object_mode)
    mode_options = [opt for opt in mode.options if not opt.volatile]
    for option in mode_options:
        try:
            validate_option_requirements(option, values)
        except ValueError:
            continue  # requirements not satisfied, skip this option
        if option.symmetric:
            value = prompt_object_option_value(option, object_mode)
        else:
            value = []
            for installation_set in range(n_sets):
                default = value[-1] if value else ''
                value.append(
                    prompt_object_option_value(
                        option=option,
                        mode=object_mode,
                        installation_set=installation_set,
                        default=default,
                    ))
            value = tuple(value)
        values[option] = value
    values = {opt.metadata: value for opt, value in values.items()}
    return values


def prompt_object_mode():
    """Prompts user for a object mode."""
    msg = 'Choose a mode: '
    completer = ObjectModeCompleter()
    validator = ContainerValidator('mode', Modes.names())
    mode = prompt(msg, completer=completer, validator=validator)
    return mode.strip()


def prompt_object_uid(package, installation_set=None):
    """Prompts user for an object UID.

    :param index: The object index within an object list.
    """
    if installation_set is None:
        installation_set = 0
    msg = 'Select an object: '
    completer = ObjectUIDCompleter(package, installation_set)
    validator = ObjectUIDValidator()
    value = prompt(msg, completer=completer, validator=validator)
    return parse_prompt_object_uid(value.strip())


def prompt_object_option(obj):
    """Prompts user for a valid option for the given object.

    :param obj: an uhu `Object` instance.
    """
    msg = 'Choose an option: '
    options = sorted(opt.metadata for opt in obj.options if not opt.volatile)
    completer = WordCompleter(options)
    validator = ContainerValidator('option', options)
    option = prompt(msg, completer=completer, validator=validator)
    return Options.get(option.strip())


def _get_object_option_value_message(option, set_=None):
    """Retuns a message for object_option_value prompt."""
    if option.default is not None:
        default_msg = option.default
        if option.type_name == 'boolean':
            if default_msg:
                default_msg = 'Y/n'
            else:
                default_msg = 'y/N'
        msg = '{} [{}]'.format(option.verbose_name.title(), default_msg)
    else:
        msg = '{}'.format(option.verbose_name.title())
    set_msg = ''
    if set_ is not None:
        set_msg = ' (installation set {})'.format(set_)
    msg = '{}{}: '.format(msg, set_msg)
    return msg


# pylint: disable=too-many-arguments
def _prompt_object_option_value(option, msg, completer, default, validator):
    """Retuns a value for object_option_value prompt."""
    value = prompt(
        msg, completer=completer, default=default, validator=validator).strip()
    if value == '':
        return option.default
    return option.validate(value)


def _get_object_option_value_completer(option, obj=None):
    """Retuns a completer for object_option_value prompt."""
    if option.choices:
        return ObjectOptionValueCompleter(option)
    elif option.type_name == 'boolean':
        return YesNoCompleter()
    elif option.metadata == 'filename':
        return ObjectFilenameCompleter()
    elif option.metadata == 'target-type':
        return WordCompleter(obj.target_types)


def prompt_object_option_value(
        option, mode, installation_set=None, default=''):
    """Given an object and an option, prompts user for a valid value.

    :param option: an uhu `Option` instance.
    :param mode: a valid Object mode string.
    :param installation_set: an int indicating the installation set.
    :param default: a default value to be displayed as placeholder.
    """
    mode = Modes.get(mode)
    msg = _get_object_option_value_message(option, installation_set)
    completer = _get_object_option_value_completer(option, mode)
    validator = ObjectOptionValueValidator(option, mode)
    value = _prompt_object_option_value(
        option, msg, completer, default, validator)
    return value


def prompt_package_uid():
    """Prompts user for a package UID."""
    msg = 'Type a package UID: '
    validator = PackageUIDValidator()
    uid = prompt(msg, validator=validator)
    return uid.strip()


def prompt_installation_set(package, msg=None):
    """Prompts user for a valid installation set.

    :param package: A core.package.Package instance.
    :param msg: The prompt message to display to user.
    :param all_sets: If True, allow to select an empty installation set.
    """
    if package.objects.is_single():
        return None

    objects = [(index, objs) for index, objs in enumerate(package.objects)]
    indexes = [str(i) for i, _ in objects]

    msg = msg if msg is not None else 'Select an installation set: '
    completer = WordCompleter(indexes)
    validator = ContainerValidator('installation set', indexes)
    installation_set = prompt(msg, completer=completer, validator=validator)
    return int(installation_set.strip())



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import json
import os
from enum import Enum

from pkgschema import validate_metadata, ValidationError

from uhu.config import config
from uhu.utils import call, get_server_url, get_chunk_size, sign_dict
from . import http


# Utilities

class ObjectReader:  # pylint: disable=too-few-public-methods
    """Read-only object class. Used when uploading with requests."""

    def __init__(self, filename, callback=None):
        self.filename = os.path.realpath(filename)
        self.callback = callback

    def __len__(self):
        return os.path.getsize(self.filename)

    def __iter__(self):
        """Yields every single chunk."""
        chunk_size = get_chunk_size()
        with open(self.filename, 'br') as fp:
            for chunk in iter(lambda: fp.read(chunk_size), b''):
                yield chunk
                call(self.callback, 'object_read')


def dummy_object_upload(filename, url, callback=None):
    data = ObjectReader(filename, callback)
    try:
        http.put(url, data=data, sign=False)
        return ObjectUploadResult.SUCCESS
    except http.HTTPError:
        return ObjectUploadResult.FAIL


def swift_object_upload(*args, **kw):
    return dummy_object_upload(*args, **kw)


def s3_object_upload(*args, **kw):
    return dummy_object_upload(*args, **kw)


STORAGES = {
    'dummy': dummy_object_upload,
    'swift': swift_object_upload,
    's3': s3_object_upload,
}


class ObjectUploadResult(Enum):
    SUCCESS = 1
    EXISTS = 2
    FAIL = 3


class UpdateHubError(Exception):
    """Exception to be used when API is broken."""


# Push Package

def push_package(metadata, objects, callback=None):
    package_uid = upload_metadata(metadata)
    upload_objects(package_uid, objects, callback)
    finish_package(package_uid, callback)
    return package_uid


def upload_metadata(metadata):
    try:
        validate_metadata(metadata)
    except ValidationError:
        raise UpdateHubError('You have an invalid package metadata.')
    url = get_server_url('/packages')
    signature = sign_dict(metadata, config.get_private_key_path())
    payload = json.dumps(metadata, sort_keys=True)
    headers = {'UH-SIGNATURE': signature}
    try:
        response = http.post(
            url, payload=payload, json=True, headers=headers).json()
        return response['uid']
    except http.HTTPError as error:
        raise UpdateHubError('Could not upload metadata: {}'.format(error))
    except (ValueError, KeyError):
        raise UpdateHubError('Could not upload metadata: unknown error.')


def upload_object(obj, package_uid, callback=None):
    """Uploads a package object to UpdateHub server."""
    # First, check if we should upload the object
    url = get_server_url('/packages/{}/objects/{}'.format(
        package_uid, obj['sha256sum']))
    body = json.dumps({'etag': obj['md5']})
    try:
        response = http.post(url, body, json=True)
    except http.HTTPError:
        return ObjectUploadResult.FAIL

    # Object already uploaded, return EXISTS.
    if response.status_code == 200:
        call(callback, 'object_read', obj['chunks'])
        return ObjectUploadResult.EXISTS

    # Object not uploaded, try to uploaded it.
    try:
        body = response.json()
        uploader = STORAGES[body['storage']]
        url = body['url']
    except (ValueError, KeyError):
        return ObjectUploadResult.FAIL
    return uploader(obj['filename'], url, callback)


def upload_objects(package_uid, objects, callback=None):
    call(callback, 'start_package_upload', objects)
    results = [upload_object(obj, package_uid, callback) for obj in objects]
    call(callback, 'finish_package_upload')
    if ObjectUploadResult.FAIL in results:
        raise UpdateHubError(
            'Some objects has not been fully uploaded. Try again later.')


def finish_package(package_uid, callback=None):
    url = get_server_url('/packages/{}/finish'.format(package_uid))
    try:
        http.put(url)
    except http.HTTPError as error:
        raise UpdateHubError(
            'Could not finish package on server: {}'.format(error))
    finally:
        call(callback, 'push_finish', package_uid)


# Package status

def get_package_status(package_uid):
    url = get_server_url('/packages/{}'.format(package_uid))
    try:
        return http.get(url, json=True).json()['status']
    except http.HTTPError as error:
        raise UpdateHubError(error)
    except (ValueError, KeyError):
        raise UpdateHubError('Could not get package info. Try again later.')



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import configparser
import os

from .utils import get_global_config_file, get_credentials, PRIVATE_KEY_FN


MAIN_SECTION = 'settings'
AUTH_SECTION = 'auth'


class Config:
    """This is the wrapper to manage ~/.uhu configuration file."""

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super().__new__(cls)
        return cls.instance

    def __init__(self):
        self._filename = get_global_config_file()
        self._config = configparser.ConfigParser(
            default_section=MAIN_SECTION)

    def get_credentials(self):
        env_credentials = get_credentials()
        config_credentials = self.get_credentials_from_config()
        credentials = env_credentials or config_credentials
        if not credentials:
            raise ValueError('Could not find any crendentials.')
        return credentials

    def get_private_key_path(self):
        env_fn = os.environ.get(PRIVATE_KEY_FN)
        pub_fn = self.get('private_key_path', AUTH_SECTION)
        private_key = env_fn or pub_fn
        if not private_key:
            raise ValueError('Could not find any private key.')
        return private_key

    def set_credentials(self, access_id, access_secret):
        """Set server requried credentials."""
        self.set('access_id', access_id, section=AUTH_SECTION)
        self.set('access_secret', access_secret, section=AUTH_SECTION)

    def get_credentials_from_config(self):
        access = self.get('access_id', AUTH_SECTION)
        secret = self.get('access_secret', AUTH_SECTION)
        if access and secret:
            return access, secret

    def set_private_key_path(self, fn):
        if not os.path.isfile(fn):
            raise ValueError('Private key is not a valid file.')
        self.set('private_key_path', fn, section=AUTH_SECTION)

    def _read(self):
        if os.path.exists(self._filename):
            self._config.read(self._filename)
        open(self._filename, 'a').close()

    def set(self, key, value, section=None):
        """Adds a new entry on settings based on key and value."""
        self._read()

        if section is None:
            section = MAIN_SECTION
        elif not self._config.has_section(section):
            self._config.add_section(section)

        self._config.set(section, key, value)

        with open(self._filename, 'w') as fp:
            self._config.write(fp)

    def get(self, key, section=None):
        """Gets the value for the given a key."""
        self._read()
        if not section:
            section = MAIN_SECTION
        value = self._config.get(section, key, fallback=None)
        return value


config = Config()  # pylint: disable=invalid-name



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

from itertools import chain

from .object import Object
from ._options import Options

from ..utils import call, list_to_str


class ObjectsManager:

    metadata = 'objects'

    MIN_N_SETS = 1
    MAX_N_SETS = 2

    def __init__(self, n_sets=2, dump=None):
        self.n_sets = None
        self.objects = None
        if dump is None:
            self._init_empty(n_sets)
        else:
            self._init_from_dump(dump)

    def _init_empty(self, n_sets):
        self.n_sets = self._validate_n_sets(n_sets)
        self.objects = []

    def _init_from_dump(self, dump):
        sets = dump.get(self.metadata)
        if sets is None:
            raise ValueError('objects key is not present within dump')
        if not isinstance(sets, list):
            raise TypeError('objects key has an invalid value type')
        self.n_sets = self._validate_n_sets(len(sets))
        self.objects = [tuple(Object(obj) for obj in objs)
                        for objs in zip(*sets)]
        self.sort()

    def _validate_n_sets(self, n_sets):
        if n_sets < self.MIN_N_SETS or n_sets > self.MAX_N_SETS:
            error = ('It is only possible to have between '
                     '{} and {} installation sets.')
            raise ValueError(error.format(self.MIN_N_SETS, self.MAX_N_SETS))
        return n_sets

    def load(self, callback=None):
        call(callback, 'start_objects_load')
        for obj in self.all():
            obj.load(callback=callback)
        call(callback, 'finish_objects_load')

    def create(self, options):
        """Creates a new object in all installation sets."""
        normalized_options = self._normalize_create_options_values(options)
        entry = self._create_object_entry(normalized_options)
        self.objects.append(entry)
        self.sort()
        return self.objects.index(entry)

    def _normalize_create_options_values(self, options):
        """Returns a tuple of options with n_sets size."""
        normalized_options = {}
        for opt, values in options.items():
            if not isinstance(values, tuple):
                values = tuple(values for _ in range(self.n_sets))
            normalized_options[opt] = values
        return normalized_options

    def _create_object_entry(self, options):
        entry = []
        for set_index in range(self.n_sets):
            obj_options = {}
            for opt, values in options.items():
                obj_options[opt] = values[set_index]
            entry.append(Object(obj_options))
        return entry

    def get(self, obj_index, set_index):
        """Retrives an object from an given installation set."""
        try:
            return self.objects[obj_index][set_index]
        except IndexError:
            raise ValueError('Object not found')

    def update(self, obj_index, option, value, set_index=None):
        """Updates an object option value."""
        option = Options.get(option)
        if option.symmetric:
            self._update_symmetric_option(obj_index, option, value)
        else:
            self._update_asymmetric_option(obj_index, set_index, option, value)

    def _update_symmetric_option(self, obj_index, option, value):
        for obj in self.objects[obj_index]:
            obj.update(option.metadata, value)

    def _update_asymmetric_option(self, obj_index, set_index, option, value):
        if set_index is None:
            error = 'You must specify an installation set for this option'
            raise ValueError(error)
        obj = self.objects[obj_index][set_index]
        obj.update(option.metadata, value)

    def remove(self, obj_index):
        """Removes an object from all sets."""
        try:
            self.objects.pop(obj_index)
        except IndexError:
            raise ValueError('Object not found')

    def all(self):
        """Returns all objects from all sets."""
        return list(chain.from_iterable(self.objects))

    def sort(self):
        self.objects.sort(key=lambda objs: objs[0].filename)

    def is_single(self):
        """Checks if it is single mode."""
        return self.n_sets == 1

    def to_metadata(self, callback=None):
        sets = self._to_list_of_sets()
        objects = [[obj.to_metadata(callback) for obj in set_]
                   for set_ in sets]
        return {self.metadata: objects}

    def to_template(self):
        sets = self._to_list_of_sets()
        objects = [[obj.to_template() for obj in set_] for set_ in sets]
        return {self.metadata: objects}

    def to_upload(self):
        sets = self._to_list_of_sets()
        objects = [obj.to_upload() for obj in sets[0]]
        return objects

    def _to_list_of_sets(self):
        return [[objs[set_index] for objs in self.objects]
                for set_index in range(self.n_sets)]

    def __eq__(self, other):
        return self.to_metadata() == other.to_metadata()

    def __getitem__(self, set_index):
        """Returns an installation set."""
        if not isinstance(set_index, int):
            raise TypeError('Installation set index must be a integer')
        if set_index < 0 or set_index > self.n_sets:
            raise IndexError('Installation set not found')
        return [objs[set_index] for objs in self.objects]

    def __len__(self):
        return self.n_sets

    def __str__(self):
        if not self.all():
            return 'Objects: None'
        sets = self._to_list_of_sets()
        lst = [self._objects_to_str(objects) for objects in sets]
        return list_to_str('Objects', lst)

    @staticmethod
    def _objects_to_str(objects):
        """Returns a installation set string representation."""
        lst = [str(obj) for obj in objects]
        return list_to_str('Installation Set', lst)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0
"""Main UHU REPL command functions."""

from ..config import config
from ..updatehub.api import get_package_status
from ..core.utils import dump_package
from ..ui import get_callback

from . import helpers
from .helpers import prompt


# Config

@helpers.cancellable
def set_authentication():
    """Sets user access and secret keys."""
    access = prompt('UpdateHub Access Key ID: ')
    secret = prompt('UpdateHub Access Secret Key: ')
    config.set_credentials(access, secret)


@helpers.cancellable
def set_private_key():
    """Sets private key path for signing packages."""
    path = prompt('UpdateHub Private Key path: ')
    config.set_private_key_path(path)


# Product

def set_product_uid(ctx):
    """Set product UID."""
    helpers.check_arg(ctx, 'You need to pass a product id')
    ctx.package.product = ctx.arg
    ctx.prompt = helpers.set_product_prompt(ctx.arg)


# Package

def set_package_version(ctx):
    """Sets the current package version."""
    helpers.check_arg(ctx, 'You need to pass a version number')
    ctx.package.version = ctx.arg


def show_package(ctx):
    """Prints package content."""
    print(ctx.package)


def save_package(ctx):
    """Save a local package file based in current package."""
    helpers.check_arg(ctx, 'You need to pass a filename')
    dump_package(ctx.package.to_template(), ctx.arg)


# Objects

@helpers.cancellable
def add_object(ctx):
    """Add an object into the current package."""
    obj_mode = helpers.prompt_object_mode()
    options = helpers.prompt_object_options(len(ctx.package.objects), obj_mode)
    options['mode'] = obj_mode
    ctx.package.objects.create(options)


@helpers.cancellable
def remove_object(ctx):
    """Removes an object from the current package."""
    obj_index = helpers.prompt_object_uid(ctx.package)
    ctx.package.objects.remove(obj_index)


@helpers.cancellable
def edit_object(ctx):
    """Edit an object within the current package."""
    obj_index = helpers.prompt_object_uid(ctx.package, 0)
    obj = ctx.package.objects.get(obj_index=obj_index, set_index=0)
    option = helpers.prompt_object_option(obj)

    set_index = None
    if not option.symmetric:
        set_index = helpers.prompt_installation_set(ctx.package)
        obj = ctx.package.objects.get(obj_index=obj_index, set_index=set_index)
    current_value = obj[option.metadata]
    default = current_value if current_value else ''
    value = helpers.prompt_object_option_value(
        option, obj.mode, default=default)
    ctx.package.objects.update(
        obj_index, option.metadata, value, set_index=set_index)


# Transactions

def push_package(ctx):
    """Upload the current package to server."""
    helpers.check_product(ctx)
    helpers.check_version(ctx)
    callback = get_callback()
    ctx.package.uid = ctx.package.push(callback)


def package_status(ctx):
    """Get the status from a package already pushed to server."""
    helpers.check_product(ctx)
    helpers.check_arg(ctx, 'You need to pass a package id')
    print(get_package_status(ctx.arg))


# Supported hardware

@helpers.cancellable
def add_hardware(ctx):
    """Adds a hardware identifier to supported hardware list."""
    identifier = prompt('Hardware identifier: ').strip()
    if identifier:
        ctx.package.supported_hardware.add(identifier)


@helpers.cancellable
def remove_hardware(ctx):
    """Removes a hardware identifier from supported hardware list."""
    identifier = prompt('Hardware identifier: ').strip()
    if identifier:
        ctx.package.supported_hardware.remove(identifier)


def reset_hardware(ctx):
    """Removes all supported hardware identifiers."""
    ctx.package.supported_hardware.reset()



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import click

from ..config import config
from ..utils import remove_local_config
from .utils import error


@click.group(name='config')
def config_cli():
    """Configures uhu utility."""


@config_cli.command(name='init')
def init_command():
    """Sets uhu required initial configuration."""
    access = input('UpdateHub Access Key ID: ')
    secret = input('UpdateHub Systems Secret Access Key: ')
    config.set_credentials(access, secret)
    path = input('UpdateHub Private Key path: ')
    config.set_private_key_path(path)


@config_cli.command(name='set')
@click.argument('entry')
@click.argument('value')
@click.option('--section', help='Section to write the configuration')
def set_command(entry, value, section):
    """Sets the given VALUE in a configuration ENTRY."""
    config.set(entry, value, section=section)


@config_cli.command(name='get')
@click.argument('entry')
@click.option('--section', help='Section to write the configuration')
def get_command(entry, section):
    """Gets the value from a given ENTRY."""
    value = config.get(entry, section=section)
    if value:
        print(value)


@click.command('cleanup')
def cleanup_command():
    """Removes uhu local config file."""
    try:
        remove_local_config()
    except FileNotFoundError:
        error(1, 'Package file already deleted.')



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0
"""UpdateHub REPL, an interactive prompt to work with firmware updates."""

import os
import sys

from prompt_toolkit.contrib.completers import WordCompleter
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.contrib.regular_languages.completion import GrammarCompleter  # nopep8
from prompt_toolkit.contrib.regular_languages import compiler

from .. import get_version
from ..config import config
from ..core.package import Package
from ..core.utils import dump_package, load_package
from ..utils import get_local_config_file

from . import functions
from .exceptions import CancelPromptException
from .helpers import prompt, set_product_prompt


COMMANDS = {
    'auth': lambda _: functions.set_authentication(),
    'quit': lambda _: sys.exit(0),
    'show': functions.show_package,
    'save': functions.save_package,
}

GROUPS = {
    'product': {
        'use': functions.set_product_uid,
    },
    'package': {
        # package
        'version': functions.set_package_version,
        # objects
        'add': functions.add_object,
        'edit': functions.edit_object,
        'remove': functions.remove_object,
        # transactions
        'push': functions.push_package,
        'status': functions.package_status,
    },
    'hardware': {
        'add': functions.add_hardware,
        'remove': functions.remove_hardware,
        'reset': functions.reset_hardware,
    }
}


GRAMMAR = compiler.compile(r'''
(\s* (?P<group>product) \s+ (?P<product>[a-z]+) \s+ (?P<arg>.+?) \s*) |
(\s* (?P<group>product) \s+ (?P<product>[a-z]+) \s*) |

(\s* (?P<group>hardware) \s+ (?P<hardware>[a-z]+) \s+ (?P<arg>.+?) \s*) |
(\s* (?P<group>hardware) \s+ (?P<hardware>[a-z]+) \s*) |

(\s* (?P<group>package) \s+ (?P<package>[a-z-]+) \s+ (?P<arg>.+?) \s*) |
(\s* (?P<group>package) \s+ (?P<package>[a-z-]+) \s*) |

(\s* (?P<command>[a-z]+) \s*) |
(\s* (?P<command>[a-z]+) \s+ (?P<arg>.+?) \s*) |
''')


class UHURepl:
    """The main class for UpdateHub REPL."""

    completer = GrammarCompleter(GRAMMAR, {
        'command': WordCompleter(COMMANDS),
        'group': WordCompleter(GROUPS),
        'hardware': WordCompleter(GROUPS['hardware']),
        'product': WordCompleter(GROUPS['product']),
        'package': WordCompleter(GROUPS['package'])
    })

    def __init__(self, package_fn=None):
        """Creates a new interactive prompt.

        The interactive prompt will work using an exisiting
        configuration file (if it exists).

        It is also able to load an existing configuration file placed
        in a non standard place if REPL is created using the
        `package_fn`.

        Finally, if there is no configuration file present in the
        working directory, neighter a package file is explicty passed,
        `UHURepl` will create a new one.

        :param package_fn: An UHU package filename.
        """
        self.local_config = get_local_config_file()

        if package_fn is not None:
            self.package = self.load_package(package_fn)
        elif os.path.exists(self.local_config):
            self.package = self.load_package(self.local_config)
        else:
            self.package = Package()

        if self.package.product:
            self.prompt = set_product_prompt(self.package.product)
        else:
            self.prompt = 'uhu> '

        self.arg = None
        self.history = InMemoryHistory()

    @staticmethod
    def load_package(fn):
        try:
            return load_package(fn)
        except ValueError as err:
            print('Error: Invalid configuration file: {}'.format(err))
            sys.exit(1)

    def repl(self):
        """Starts a new interactive prompt."""
        print('UpdateHub Utils {}'.format(get_version()))
        while True:
            try:
                expression = prompt(
                    self.prompt,
                    completer=self.completer,
                    history=self.history
                )
            except CancelPromptException:
                sys.exit(1)  # User has typed Ctrl C
            try:
                command = self.get_command(expression)
            except TypeError:  # Invalid expression
                print('ERROR: Invalid command')
                continue
            except ValueError:  # Empty prompt
                continue
            else:
                self.run_command(command)

    def get_command(self, expression):
        """Given an expression, returns a valid command.

        :param expression: The full expression typed by user on
                           prompt.
        """
        expression = GRAMMAR.match(expression)

        if expression is None:
            raise TypeError

        vars_ = expression.variables()
        cmd, cmd_group = vars_.get('command'), vars_.get('group')
        self.arg = vars_.get('arg')

        if cmd is not None:
            command = COMMANDS.get(cmd)
        elif cmd_group is not None:
            group, cmd = GROUPS.get(cmd_group), vars_.get(cmd_group)
            command = group.get(cmd)
        else:
            raise ValueError('Invalid command')
        return command

    def run_command(self, command):
        """Executes an given command.

        If command runs successfully, persists the configuration state
        into a file. Otherwise, shows the error message to user.
        """
        if command is None:
            print('Invalid command')
            return
        try:
            command(self)
        except Exception as err:  # pylint: disable=broad-except
            print('\033[91mError:\033[0m {}'.format(err))
        else:  # save package in every successful command
            dump_package(self.package.to_template(), self.local_config)


def repl(package):
    """Instantiates a new UHURepl.

    Before creating a new instance, checks if server authentication is
    set. If not, prompts user for its credentials.
    """
    try:
        config.get_credentials()
    except ValueError:
        functions.set_authentication()
    try:
        config.get_private_key_path()
    except ValueError:
        functions.set_private_key()
    return UHURepl(package).repl()



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import click

from .utils import open_package


@click.group(name='product')
def product_cli():
    """Product related commands."""


@product_cli.command(name='use')
@click.argument('uid')
def use_command(uid):
    """Sets the product."""
    with open_package() as package:
        package.product = uid
